#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "bytesource.h"
#include "dehuff.h"
#include "input.h"

void read_huffman_tables(huffman_tables_t* dst, input_func_t* input_func, void* userdata)
{
	size_t len = read_uint16(input_func, userdata);
	assert(len > 2);
	len -= 2;

	// TODO: check for NULL return
	uint8_t* buf = (uint8_t*)malloc(len);
	uint8_t* inptr = buf;
	reliable_read(input_func, userdata, buf, len);
	
	while (len > 0) {
		unsigned char tc_th = *inptr++;
		--len;
		unsigned table_class = tc_th >> 4;
		unsigned table_dest = tc_th & 0x0f;
		
		if (table_class != 0 && table_class != 1) {
			fprintf(stderr, "Error: Unknown table class %u\n", table_class);
			exit(1);
		}
		if (table_dest >= 4) {
			fprintf(stderr, "Error: Unknown table destination %u\n", table_dest);
			exit(1);
		}

		struct huffman_table* tbl = &((*dst)[table_class][table_dest]);
		if (len < 16) {
			fprintf(stderr, "Short read for num_codes\n");
			exit(1);
		}
		for (unsigned i = 0; i < 16; ++i) {
			tbl->num_codes[i] = *inptr++;
			--len;
		}
		for (unsigned i = 0, k = 0; i < 16; ++i) {
			if (len < tbl->num_codes[i]) {
				fprintf(stderr, "Short read for codes\n");
				exit(1);
			}
			for (unsigned j = 0; j < tbl->num_codes[i]; ++j, ++k) {
				tbl->codes[k] = *inptr++;
				--len;
			}
		}

		// Generate derived tables

		// generate_size_table (figure C.1)
		{
			unsigned k = 0;
			for (unsigned i = 0; i < 16; ++i) {
				for (unsigned j = 0; j < tbl->num_codes[i]; ++j, ++k) {
					tbl->huffsize[k] = i + 1;
				}
			}
			tbl->huffsize[k] = 0;
		}

		// generate_code_table (figure C.2)
		unsigned si = tbl->huffsize[0];
		for (unsigned i = 0, j = 0;; ) {
			tbl->huffcode[i++] = j++;
			if (tbl->huffsize[i] == si) {
				continue;
			}
			if (tbl->huffsize[i] == 0) {
				break;
			}

			do {
				j <<= 1;
			} while (++si != tbl->huffsize[i]);
		}

		// decoder_tables (figure F.15)
		for (unsigned i = 0, j = 0; i < 16; ++i) {
			if (tbl->num_codes[i] == 0) {
				tbl->maxcode[i] = -1;
				continue;
			}

			tbl->valptr[i] = j;
			tbl->mincode[i] = tbl->huffcode[j];
			j += tbl->num_codes[i];
			tbl->maxcode[i] = tbl->huffcode[j - 1];
		}

		// Generate the lookup tables
		for (unsigned i = 0; i < DEHUF_TABLE_SIZE; ++i) {
			tbl->lookup_table_codes[i] = DEHUF_SLOW_PATH;
			tbl->lookup_table_length[i] = DEHUF_SLOW_PATH;
		}
	
		unsigned k = 0;	
		for (unsigned i = 0; i < 16; ++i) {
			for (unsigned j = 0; j < tbl->num_codes[i]; ++j, ++k) {
				const unsigned code = tbl->huffcode[k];
				const unsigned length = tbl->huffsize[k];
				if (length > DEHUF_TABLE_BITS) {
					continue;
				}
				const unsigned prefix_min = code << (DEHUF_TABLE_BITS - length);
				const unsigned prefix_max = ((code + 1) << (DEHUF_TABLE_BITS - length));

				for (unsigned elem = prefix_min; elem < prefix_max; ++elem) {
					assert(tbl->lookup_table_codes[elem] == DEHUF_SLOW_PATH);
					assert(tbl->lookup_table_length[elem] == DEHUF_SLOW_PATH);
					tbl->lookup_table_codes[elem] = tbl->codes[k];
					tbl->lookup_table_length[elem] = length;
				}
			}
		}	
	}

	free(buf);
}

unsigned read_huffman_symbol_slow_path(const struct huffman_table* table,
                                       struct bit_source* source) {
	possibly_refill(source, 1);
	unsigned code = read_bits(source, 1);
	int i = 0;

	while (table->maxcode[i] == -1 || code > (unsigned)table->maxcode[i]) {
		possibly_refill(source, 1);
		code = (code << 1) | read_bits(source, 1);
		++i;

		if (i > 15) {
			fprintf(stderr, "Error in Huffman decoding: Too long code (%x)\n", code);
			exit(1);
		}
	}

	return table->codes[table->valptr[i] + code - table->mincode[i]];
}
