#ifndef _DEHUFF_H
#define _DEHUFF_H 1

#include <stddef.h>
#include <stdint.h>
#include <sys/types.h>

#include "bytesource.h"
#include "bitsource.h"

// About 99% of all Huffman codes are <= 8 bits long (see codelen.txt),
// and it's what libjpeg uses. Thus, it seems like a reasonable size.
#define DEHUF_TABLE_BITS 8
#define DEHUF_TABLE_SIZE (1 << DEHUF_TABLE_BITS)
static const int DEHUF_SLOW_PATH = -1;

struct huffman_table {
	unsigned num_codes[17];     // BITS
	unsigned char codes[256];   // HUFFVAL
	
	// Derived values.
	unsigned huffsize[256];
	unsigned huffcode[256];
	int maxcode[16];
	int mincode[16];
	unsigned valptr[16];

	// Lookup table for fast decoding; given eight bits,
	// return the symbol and length in bits. For longer codes,
	// DEHUF_SLOW_PATH is returned.

	// Note that the codes we return are 8-bit, but the type of
	// the lookup tables is int to avoid extra zero extending. 
	int lookup_table_codes[DEHUF_TABLE_SIZE]; 
	int lookup_table_length[DEHUF_TABLE_SIZE]; 
};

enum coefficient_class {
	DC_CLASS = 0,
	AC_CLASS,
	NUM_COEFF_CLASSES
};
typedef struct huffman_table huffman_tables_t[NUM_COEFF_CLASSES][4];

// Read Huffman tables from a stream, and compute the derived values.
void read_huffman_tables(huffman_tables_t* dst, input_func_t* input_func, void* userdata);

unsigned read_huffman_symbol_slow_path(const struct huffman_table* table,
                                       struct bit_source* source);

static inline unsigned read_huffman_symbol(const struct huffman_table* table,
                                           struct bit_source* source)
{
	// FIXME: We can read past the end of the stream here in some edge
	// cases. We need to define some guarantees in the layers above.
	possibly_refill(source, DEHUF_TABLE_BITS);
	unsigned lookup = peek_bits(source, DEHUF_TABLE_BITS);
	int code = table->lookup_table_codes[lookup];
	int length = table->lookup_table_length[lookup];

	if (code == DEHUF_SLOW_PATH) {
		return read_huffman_symbol_slow_path(table, source);
	}

	read_bits(source, length);
	return code;
}

// procedure EXTEND (figure F.12)

// Fast lookup table for (1 << (bits - 1)).
// The table actually helps, since the load can go in parallel with the shift
// operation below.
static const int bit_thresholds[16] = {
	0, 1 << 0, 1 << 1, 1 << 2, 1 << 3, 1 << 4, 1 << 5, 1 << 6, 1 << 7, 1 << 8, 1 << 9, 1 << 10, 1 << 11, 1 << 12, 1 << 13, 1 << 14
};

static inline unsigned extend(int val, int bits)
{
	if (val < bit_thresholds[bits]) {
		return val + (-1 << bits) + 1;
	} else {
		return val;
	}
}

#endif /* !defined(_DEHUFF_H) */
