#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "bytesource.h"
#include "choice.h"
#include "dehuff.h"
#include "idct.h"
#include "input.h"
#include "zigzag.h"

struct jpeg_image {
	unsigned precision;
	unsigned width, height;
	unsigned num_components;
	unsigned hsample[256], vsample[256], qtable[256];
	uint32_t qvalues[256][DCTSIZE2];
	void* idct_data[256];
};

ssize_t stdio_read(void* userdata, uint8_t* buf, size_t count) 
{
	return fread(buf, 1, count, (FILE*)userdata);
}

void read_dqt(struct byte_source* source, struct jpeg_image* image)
{
	unsigned len = read_uint16(byte_source_input_func, source);
	assert(len >= 67);
	uint8_t precision_table = read_uint8(byte_source_input_func, source);
	int precision = precision_table >> 4;  // 0 = 8 bits, otherwise 16 bits.
	int table = precision_table & 0x0f;

	if (image->idct_data[table] != NULL) {
		idct_choice_free(image->idct_data[table]);
	}

	if (precision != 0) {
		assert(len == 131);
		fprintf(stderr, "Quantization table %u: 16 bits/entry\n", table);
	} else {
		assert(len == 67);
		fprintf(stderr, "Quantization table %u: 8 bits/entry\n", table);
	}
	
	for (unsigned i = 0; i < 64; ++i) {
		if (precision != 0) {
			image->qvalues[table][unzigzag[i]] =
				read_uint16(byte_source_input_func, source);
		} else {
			image->qvalues[table][unzigzag[i]] =
				read_uint8(byte_source_input_func, source);
		}	
	}

	image->idct_data[table] = idct_choice_alloc(image->qvalues[table]);
}

void read_sof(struct byte_source* source, struct jpeg_image* image)
{
	unsigned len = read_uint16(byte_source_input_func, source);
	assert(len >= 8);
	image->precision = read_uint8(byte_source_input_func, source);
	assert(image->precision == 8);
	image->height = read_uint16(byte_source_input_func, source);
	image->width = read_uint16(byte_source_input_func, source);
	image->num_components = read_uint8(byte_source_input_func, source);
	len -= 8;

	fprintf(stderr, "%u-bit %ux%u JPEG with %u components\n",
	        image->precision, image->width, image->height, image->num_components);

	for (unsigned i = 0; i < image->num_components; ++i) {
		assert(len >= 3);
		unsigned c = read_uint8(byte_source_input_func, source);
		unsigned sampling_factors = read_uint8(byte_source_input_func, source);
		image->hsample[c] = sampling_factors >> 4;
		image->vsample[c] = sampling_factors & 0x0f;
		image->qtable[c] = read_uint8(byte_source_input_func, source);
		len -= 3;

		fprintf(stderr, "Component %u: sampling factors %u x %x, quantization table %u\n",
			c, image->hsample[c], image->vsample[c], image->qtable[c]);
	}
}

void read_scan(struct byte_source* source, struct jpeg_image* image, huffman_tables_t* tables)
{
	unsigned len = read_uint16(byte_source_input_func, source);
	assert(len >= 2);
	len -= 2;

	assert(len >= 1);
	unsigned num_components = read_uint8(byte_source_input_func, source);
	--len;

	unsigned component_num[256];
	unsigned dc_huffman_table[256], ac_huffman_table[256];
	unsigned ss, se, ah_al;
	int last_dc[256];

	for (unsigned i = 0; i < num_components; ++i) {
		unsigned char td_ta;
		assert(len >= 2);
		component_num[i] = read_uint8(byte_source_input_func, source);
		td_ta = read_uint8(byte_source_input_func, source);
		len -= 2;
		dc_huffman_table[i] = td_ta >> 4;
		ac_huffman_table[i] = td_ta & 0x0f;
		last_dc[i] = 0;
	}

	assert(len >= 3);
	ss = read_uint8(byte_source_input_func, source);
	se = read_uint8(byte_source_input_func, source);
	ah_al = read_uint8(byte_source_input_func, source);
	len -= 3;

	if (len != 0) {
		fprintf(stderr, "Error: %u unused bytes at end of SOS segment\n", len);
	}

	struct bit_source bits;
	init_bit_source(&bits, byte_source_input_func, source);
		
	// Some debugging code.
	assert(image->width % 8 == 0);
	int y_stride = image->width;
	uint8_t* y_row_data = (uint8_t*)malloc(y_stride * DCTSIZE * image->vsample[1]);
	assert(y_row_data != NULL);
	int xb = 0;

	for ( ;; ) {
		for (unsigned c = 0; c < num_components; ++c) {
			unsigned cn = component_num[c];
			assert(image->idct_data[image->qtable[cn]] != NULL);
			
			for (unsigned local_yb = 0; local_yb < image->vsample[cn]; ++local_yb) {
				for (unsigned local_xb = 0; local_xb < image->hsample[cn]; ++local_xb) {
					const struct huffman_table* dc_table = &((*tables)[DC_CLASS][dc_huffman_table[c]]);
					const struct huffman_table* ac_table = &((*tables)[AC_CLASS][ac_huffman_table[c]]);

					// decode DC component
					unsigned dc_category = read_huffman_symbol(dc_table, &bits);
					possibly_refill(&bits, dc_category);
					last_dc[c] += extend(read_bits(&bits, dc_category), dc_category);
					
					int16_t coeff[DCTSIZE2] = { 0 };
					coeff[0] = last_dc[c];

					// decode AC components
					for (unsigned i = 1; i < DCTSIZE2; ++i) {
						unsigned rs = read_huffman_symbol(ac_table, &bits);
						unsigned r = rs >> 4;
						unsigned s = rs & 0xf;

						if (rs == 0x00) {
							/* end of block */
							break;
						}
						if (rs == 0xf0) {
							/* 16 zero coefficients */
							i += 15;
							continue;
						}
						i += r;

						possibly_refill(&bits, s);
						coeff[unzigzag[i]] = extend(read_bits(&bits, s), s);
					}
			
					uint8_t pixdata[DCTSIZE2];	
					idct_choice(coeff, image->idct_data[image->qtable[cn]], pixdata);

					if (cn != 1) {
						continue;
					}

					for (int y = 0; y < DCTSIZE; ++y) {
						memcpy(y_row_data + (local_yb * DCTSIZE + y) * y_stride + (xb + local_xb) * DCTSIZE,
						       pixdata + y * DCTSIZE,
						       DCTSIZE);
					}
				}
			}

			if (cn != 1) {
				continue;
			}

			xb += image->hsample[cn];
			if (xb * DCTSIZE == y_stride) {
				fwrite(y_row_data, y_stride * DCTSIZE * image->vsample[cn], 1, stdout);
				xb = 0;
			}
		}
	}
	if (len != 0) {
		fprintf(stderr, "Error: %u unused bytes at end of SOS segment\n", len);
	}
}
			
void skip_segment(struct byte_source* source)
{
	uint8_t buf[4096];
	for ( ;; ) {
		ssize_t ret = byte_source_input_func(source, buf, 4096);
		if (ret == -1) {
			fprintf(stderr, "Input error!\n");
			exit(1);
		}
		if (ret == 0) {
			return;
		}
	}
}
	
int main(void)
{
	struct jpeg_image jpeg;
	memset(&jpeg, 0, sizeof(jpeg));
	init_choices();

	struct byte_source source;
	init_byte_source(&source, stdio_read, stdin);

	huffman_tables_t tables;

	for ( ;; ) {
		uint8_t m2 = byte_source_read_marker(&source);
		assert(m2 != 0x00);

		fprintf(stderr, "Marker 0x%02x, at position %ld\n", m2, ftell(stdin) - source.bytes_available);

		switch (m2) {
		case 0xe0:
		case 0xe1:
		case 0xe2:
		case 0xe3:
		case 0xe4:
		case 0xe5:
		case 0xe6:
		case 0xe7:
		case 0xe8:
		case 0xe9:
		case 0xea:
		case 0xeb:
		case 0xec:
		case 0xed:
		case 0xee:
		case 0xef:
			/* APP0 through APPF */
		case 0xfc:
			/* some EXIF stuff */
		case 0xfe:
			/* comment */
		case 0xff:
			/* ignore */
			skip_segment(&source);
			break;
		case 0xdb:
			/* DQT */
			read_dqt(&source, &jpeg);
			break;
		case 0xc0:
			/* SOF0 (baseline DCT, Huffman encoded) */
			read_sof(&source, &jpeg);
			break;
		case 0xd8:
			/* SOI */
			break;
		case 0xc4:
			/* DHT (define Huffman tables) */
			read_huffman_tables(&tables, byte_source_input_func, &source);
			break;
		case 0xda:
			/* SOS (start of scan) */
			read_scan(&source, &jpeg, &tables);
			break;
		default:
			fprintf(stderr, "Error: Unknown marker 0x%02x\n", m2);
			exit(1);
		}
	}
}
